#!/usr/bin/env python

import sys
import time
import random

try:
	import duptreap
except ImportError:
	means = 'dict'
else:
	means = 'treap'

class blocksize_class:
	def __init__(self, size):
		self.size = size
		self.total_size = 0L
		self.total_time = 0L

	def add_duration(self, duration, length):
		self.total_time += duration
		self.total_size += length
		self.average = float(self.total_time) / self.total_size

	def get_average(self):
		return self.average

	def __str__(self):
		s = '%d, %f, %f' % (self.size, self.total_time, self.average)
		return s

# Initially we get a little and put a lot
# After we get established, we get all the time, and put a lot

class common:
	def generate_random_blocksize(self):
		width = self.highest - self.lowest
		guess = int(random.random()*width) + self.lowest
		self.decrement()
		return guess

	def decrement(self):
		self.probability_of_guess = max(self.probability_of_guess - self.probability_of_guess_decrement, 0.001)

if means == 'dict':
	sys.stderr.write('%s: Warning: Using a dict+list, not a dict+treap: storage requirements will increase slowly, but without bound\n' % \
		sys.argv[0])
	# This implementation is:
	#    O(n) on get
	#    O(1) (assuming good hashing) on put
	#    Grows without bound, albeit slowly
	class blocksize_optimizer(common):
		def __init__(self, lowest, highest, probability_of_guess_decrement):
			self.lowest = lowest
			self.highest = highest
			# initially the probability of a guess is 100%, but need not be
			self.probability_of_guess = 1.0
			self.probability_of_guess_decrement = probability_of_guess_decrement
			self.blocksize_average = {}
			# nice for debugging
			#random.seed(0)

		def put(self, length, duration):
			if not self.blocksize_average.has_key(length):
				self.blocksize_average[length] = blocksize_class(length)
			self.blocksize_average[length].add_duration(duration, length)

		def get(self):
			if random.random() < self.probability_of_guess:
				return self.generate_random_blocksize()
			else:
				lst = []
				for key, value in self.blocksize_average.items():
					lst.append((value.get_average(), key))
				if lst:
					minimum = min(lst)
					self.decrement()
					best = minimum[1]
					return best
				else:
					return self.generate_random_blocksize()
elif means == 'treap':
	# This implementation is:
	#    O(logn) on put
	#    O(logn) on get
	#
	# Because after getting established we put And get on almost every operation, this is a performance win
	class blocksize_optimizer(common):
		def __init__(self, lowest, highest, probability_of_guess_decrement):
			self.lowest = lowest
			self.highest = highest
			# initially the probability of a guess is 100%, but need not be
			self.probability_of_guess = 1.0
			self.probability_of_guess_decrement = probability_of_guess_decrement
			self.blocksizes_by_size = {}
			self.blocksizes_by_throughput = duptreap.duptreap()
			self.blocksizes_in_limbo = {}
			self.num_blocksizes = 0
			# nice for debugging
			#random.seed(0)

		def put(self, length, duration):
			if length in self.blocksizes_by_size:
				# We already have this block size in blocksizes_by_size - that's strange
				sys.stderr.write('%s: Internal error: blocksize already in blocksizes_by_size\n' % sys.argv[0])
				sys.exit(1)
			elif length in self.blocksizes_in_limbo:
				# We've seen this blocksize before.  In fact, we passed it back to the caller, expecting to see it again later.
				# Update the blocksize status with the new data.
				self.blocksizes_in_limbo[length].add_duration(duration, length)
				# Put it back into blocksizes_by_size and blocksizes_by_throughput.
				self.blocksizes_by_size[length] = self.blocksizes_in_limbo[length]
				self.blocksizes_by_throughput[self.blocksizes_in_limbo[length].get_average()] = self.blocksizes_in_limbo[length]
				# Delete the blocksize from limbo
				del self.blocksizes_in_limbo[length]
			else:
				# We do not yet have this blocksize anywhere.
				# Update it with the new data.
				# Add it to the dictionary.
				# Add it to the treap.
				# Do not add it to limbo (yet).
				bs = blocksize_class(length)
				bs.add_duration(duration, length)
				self.blocksizes_by_size[length] = bs
				self.blocksizes_by_throughput[bs.get_average()] = bs
			self.num_blocksizes += 1
			if self.num_blocksizes >= 1000:
				# we have too many blocksizes to worry about them all - delete the one that was the slowest
				rate, best_blocksize = self.blocksizes_by_throughput.remove_min()
				del self.blocksizes_by_size[best_blocksize.size]
				self.num_blocksizes -= 1
			#print >> sys.stderr, 'self.num_blocksizes is', self.num_blocksizes, len(self.blocksizes_by_size), len(self.blocksizes_by_throughput)

		def get(self):
			if random.random() < self.probability_of_guess:
				# Get a new blocksize.
				# Loop until we generate a new blocksize.
				while True:
					new_blocksize = self.generate_random_blocksize()
					if new_blocksize not in self.blocksizes_by_size and new_blocksize not in self.blocksizes_in_limbo:
						break
				return new_blocksize
			else:
				# Reuse the best blocksize.
				# 1) Extract the blocksize with the best rate
				# 2) Delete it from blocksizes_by_size
				# 3) Stash it in "limbo" for later use - it'll probably be coming back again
				rate, best_blocksize = self.blocksizes_by_throughput.remove_max()
				#print >> sys.stderr, best_blocksize
				del self.blocksizes_by_size[best_blocksize.size]
				self.blocksizes_in_limbo[best_blocksize.size] = best_blocksize
				self.num_blocksizes -= 1
				return best_blocksize.size
else:
	sys.stderr.write('%s: Illegal value of means: %s\n' % (sys.argv[0], means))
	sys.exit(1)

